from jsmin import jsmin
import os

static_path = os.path.join(os.path.dirname(__file__), 'static')
asset_files = ['eventsource.js', 'reconnecting-websocket.js',
    'presence.js', 'push.js']


def generate_bundle(files=None, host=None, root=static_path):
    if not files:
        files = asset_files
    content = ""
    for filename in files:
        with open(os.path.join(root, filename)) as f:
            content += f.read()
    if host:
        content += "tornadopush.init('%s');" % host
    return jsmin(content)


def generate_authentication_code(token):
    return "tornadopush.authentify('%s');" % str(token)


def generate_tokenized_bundle(token, files, **kwargs):
    content = generate_bundle(files, **kwargs)
    content += generate_authentication_code(token)
    return content


import tornado.web
import tornado.websocket
import tornado.httpclient
from .base import *
from .queue import QueueMixin
from .presence import PresenceMixin
import json


class BaseMixin(PresenceMixin, ChannelMixin, AuthentifyMixin, QueueMixin):
    @property
    def presence_enabled(self):
        return self.get_argument('presence', None) == '1'

    def open(self, channel=None):
        self.channel = channel
        self.authentify(self.get_argument('token', None))

    def on_authorized(self):
        AuthentifyMixin.on_authorized(self)
        if self.presence_enabled:
            self.notify_join(self.channel)
        self.subscribe(self.channel_key)

    def on_event_message(self, msg):
        event, target, data = msg.split(';', 2)
        if target and target != self.user_id:
            logger.debug('Ignored event: %s' % event)
            return
        logger.debug('Event received: %s' % event)
        self.handle_event(event, data)

    on_queue_message = on_event_message

    def on_message(self, message):
        if self.application.settings['publish_method'] == 'queue':
            self.publish_queue_message(self.channel, message)
        elif self.application.settings['publish_method'] == 'webhook':
            webhook_url = self.application.settings['webhook_url']
            if not webhook_url:
                logger.error('Missing webhook url!')
            logger.debug("Message from client: %s" % message)
            http_client = tornado.httpclient.AsyncHTTPClient()
            http_client.fetch(webhook_url,
                method='POST',
                headers={"Content-Type": "application/json"},
                body=json.dumps({
                    "channel": self.channel,
                    "user_id": self.user_id,
                    "message": message}))

    def handle_event(self, event, data):
        raise NotImplementedError

    def do_close(self, callback=None):
        self.notify_leave_all()
        self.close_queue_client(callback)


class SSEHandler(BaseMixin, tornado.web.RequestHandler):
    def __init__(self, *args, **kwargs):
        tornado.web.RequestHandler.__init__(self, *args, **kwargs)
        self.presence_channels = {}
        self.auth_done = False

    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', 'Content-Type')

    @tornado.web.asynchronous
    def get(self, channel=None):
        self.set_header('Content-Type','text/event-stream; charset=utf-8')
        self.set_header('Cache-Control','no-cache')
        self.set_header('Connection','keep-alive')
        self.open(channel)

    @tornado.web.asynchronous
    def post(self, channel=None):
        self.channel = channel
        def on_auth():
            self.on_message(self.request.body)
            self.finish()
        self.authentify(self.get_argument('token', None), callback=on_auth)

    def options(self, *args, **kwargs):
        self.finish()

    def on_authorized(self):
        self.flush() # send headers
        BaseMixin.on_authorized(self)

    def on_queue_disconnect(self):
        self.finish()

    def handle_event(self, event, data):
        self.write("event: {0}\ndata: {1}\n\n".format(event, data))
        self.flush()

    def handle_presence(self, msg):
        self.handle_event("__presence__", msg)

    def on_connection_close(self):
        self.do_close(lambda: super(SSEHandler, self).on_connection_close())


class WebSocketHandler(BaseMixin, WebSocketMixin, tornado.websocket.WebSocketHandler):
    def __init__(self, *args, **kwargs):
        tornado.websocket.WebSocketHandler.__init__(self, *args, **kwargs)
        self.presence_channels = {}
        self.auth_done = False

    def on_authorized(self):
        BaseMixin.on_authorized(self)
        self.start_heartbeat()

    def on_queue_disconnect(self):
        self.close()
        self.on_close()

    def handle_event(self, event, data):
        try:
            self.write_message("event:%s;%s" % (event, data))
        except tornado.websocket.WebSocketClosedError:
            pass

    def on_message(self, message):
        if not self.auth_done:
            return
        if message.startswith('presence:'):
            self.on_presence_message(message[9:])
        elif message.startswith('message:'):
            BaseMixin.on_message(self, message[8:])

    def on_close(self):
        self.do_close()



from tornado.websocket import WebSocketHandler, WebSocketClosedError
from .base import *


class User(object):
    def __init__(self, uid, meta=None, handler=None):
        self.uid = uid
        self.meta = meta or '{}'
        self.handlers = []
        if handler:
            self.handlers.append(handler)

    def send(self, msg):
        for handler in self.handlers:
            if handler.ws_connection:
                handler.handle_presence(msg)


class Channel(object):
    def __init__(self, name):
        self.name = name
        self.users = {}

    def broadcast(self, msg, ignore_user=None):
        for user in self.users.itervalues():
            if not ignore_user or user.uid != ignore_user:
                user.send("%s:%s" % (self.name, msg))

    def broadcast_msg(self, from_uid, msg):
        self.broadcast('bcast:%s;%s' % (from_uid, msg), from_uid)
        logger.debug('Presence: broadcast from %s on #%s: %s' % (from_uid, self.name, msg))

    def join(self, uid, handler, meta=None):
        if uid not in self.users:
            self.users[uid] = User(uid, meta, handler)
            self.broadcast("+:%s;%s" % (uid, self.users[uid].meta), uid)
            logger.debug('Presence: %s joined #%s' % (uid, self.name))
        else:
            self.users[uid].handlers.append(handler)
            if meta:
                self.set_meta(uid, meta)

    def set_meta(self, uid, meta):
        self.users[uid].meta = meta
        self.broadcast("meta:%s;%s" % (uid, meta), uid)
        logger.debug('Presence: meta for %s on #%s: %s' % (uid, self.name, meta))

    def leave(self, uid, handler):
        if uid not in self.users:
            return
        user = self.users[uid]
        user.handlers.remove(handler)
        if not user.handlers:
            del self.users[uid]
            self.broadcast("-:%s" % uid, uid)
            logger.debug('Presence: %s left #%s' % (uid, self.name))

    def send_user_list(self, handler, ignore_user=None):
        for user in self.users.itervalues():
            if not ignore_user or user.uid != ignore_user:
                handler.handle_presence("%s:+:%s;%s" % (self.name, user.uid, user.meta))


CHANNELS = {}
def get_channel(name):
    if name not in CHANNELS:
        CHANNELS[name] = Channel(name)
    return CHANNELS[name]


class PresenceMixin(object):
    def notify_join(self, channel, meta=None):
        if channel not in self.presence_channels:
            if self.user_channels and channel not in self.user_channels:
                return None
            self.presence_channels[channel] = get_channel(channel)
            self.presence_channels[channel].join(self.user_id, self, meta)
            self.send_presence_user_list(channel)
        return self.presence_channels[channel]

    def notify_leave(self, channel):
        if channel in self.presence_channels:
            self.presence_channels[channel].leave(self.user_id, self)
            del self.presence_channels[channel]

    def notify_leave_all(self):
        for channel in self.presence_channels.itervalues():
            channel.leave(self.user_id, self)
        self.presence_channels = {}

    def send_presence_user_list(self, channel):
        ch = self.notify_join(channel)
        if ch:
            ch.send_user_list(self, self.user_id)

    def on_presence_message(self, msg):
        channel, action, data = msg.split(':', 2)
        if action == 'join':
            self.notify_join(channel, data)
        elif action == 'leave':
            self.notify_leave(channel)
        elif action == 'bcast':
            ch = self.notify_join(channel)
            if ch:
                ch.broadcast_msg(self.user_id, data)
        elif action == 'meta':
            ch = self.notify_join(channel)
            if ch:
                ch.set_meta(self.user_id, data)
        elif action == 'users':
            self.send_presence_user_list(channel)

    def handle_presence(self, msg):
        try:
            self.write_message("presence:%s" % msg)
        except WebSocketClosedError:
            pass


class PresenceHandler(PresenceMixin, AuthentifyMixin, WebSocketMixin, WebSocketHandler):

    def __init__(self, *args, **kwargs):
        WebSocketHandler.__init__(self, *args, **kwargs)
        self.presence_channels = {}
        self.channel = None
        self.auth_done = False

    def open(self):
        self.authentify(self.get_argument('token', None))

    def on_authorized(self):
        AuthentifyMixin.on_authorized(self)
        self.start_heartbeat()

    def on_message(self, message):
        if not self.auth_done:
            return
        if message.startswith('presence:'):
            self.on_presence_message(message[9:])

    def on_close(self):
        self.notify_leave_all()


import tornado.ioloop
import tornado.httpserver
import tornado.web
import os
import tempfile
from itsdangerous import URLSafeTimedSerializer
from .handlers import SSEHandler, WebSocketHandler
from .presence import PresenceHandler
from .base import logger
from .queue import queue_providers
from tornadopush.assets import generate_bundle, generate_authentication_code


class JsClientHandler(tornado.web.RequestHandler):
    def get(self):
        self.set_header('Content-Type', 'text/javascript')
        token = self.get_argument('token', None)
        if not token:
            content = self.generate_default_bundle()
        else:
            content = self.generate_tokenized_bundle(token)
        self.write(content)
        self.finish()

    def generate_default_bundle(self):
        tmpfilename = os.path.join(self.application.settings['assets_cache_path'], 'bundle.js')
        if not os.path.exists(tmpfilename):
            content = generate_bundle(host=self.request.host)
            with open(tmpfilename, 'w') as f:
                f.write(content)
            logger.debug('Bundle file written to %s' % tmpfilename)
            return content
        with open(tmpfilename) as f:
            return f.read()

    def generate_tokenized_bundle(self, token):
        content = self.generate_default_bundle()
        content += generate_authentication_code(token)
        return content


class Application(tornado.web.Application):
    default_settings = {
        'queue': 'tornado_redis',
        'redis_host': 'localhost',
        'redis_port': 6379,
        'redis_password': None,
        'redis_db': None,
        'token_secret': 'tornadopush',
        'auth_enabled': False,
        'channel_format': '%s',
        'presence_channel': 'presence:%s',
        'presence_users_key': 'presence_users:%s',
        'token_max_age': 3600,
        'allowed_channels': None,
        'default_channel': 'tornado',
        'publish_method': None,
        'webhook_url': None,
        'assets_cache_path': None
    }

    def __init__(self, default_host="", transforms=None, **settings):
        static_path = os.path.join(os.path.dirname(__file__), '../static')
        handlers = [
            (r"/tornadopush.js", JsClientHandler),
            (r"/static/(.*)", tornado.web.StaticFileHandler, {"path": static_path}),
            (r"/channel/([a-zA-Z_0-9]+)", WebSocketHandler),
            (r"/channel/([a-zA-Z_0-9]+)\.sse", SSEHandler),
            (r"/presence", PresenceHandler)]
        super(Application, self).__init__(handlers, default_host, transforms,
            **dict(self.default_settings, **settings))
        self.token_serializer = URLSafeTimedSerializer(self.settings["token_secret"])
        self.queue_provider = queue_providers[self.settings['queue']](self)
        logger.debug('Using queue provider: %s' % self.settings['queue'])
        if not self.settings['assets_cache_path']:
            self.settings['assets_cache_path'] = tempfile.mkdtemp()


def create_server(port=8888, **settings):
    server = tornado.httpserver.HTTPServer(Application(**settings))
    server.listen(port)
    return server


def start_server(port=8888, **settings):
    server = create_server(port, **settings)
    tornado.ioloop.IOLoop.instance().start()



from .server import start_server, logger
import logging
import yaml
import argparse
import os
import signal
import tornado.ioloop


is_closing = False
logging.basicConfig()


def signal_handler(signum, frame):
    global is_closing
    print "Closing..."
    is_closing = True


def try_exit():
    if is_closing:
        tornado.ioloop.IOLoop.instance().stop()


argparser = argparse.ArgumentParser(prog='tornadopush',
    description='Start tornadopush server')
argparser.add_argument('-p', '--port', default=8888, type=int,
    help='Port number')
argparser.add_argument('-c', '--config', default='config.yml',
    help='Path to a YAML configuration file')
argparser.add_argument('--debug', default=False, action='store_true',
    help="Enable debug messages")
argparser.add_argument('--auth', default=False, action='store_true',
    help="Whether to make authentification mandatory")
argparser.add_argument('--secret', default='tornadopush',
    help='Secret key for the token serializer')
argparser.add_argument('--tokenttl', default=3600, type=int,
    help='Token duration in seconds')
argparser.add_argument('--default-channel', default='tornado',
    help='Default channel name')
argparser.add_argument('--assets-cache-path',
    help='Path where assets bundles will be compiled')
argparser.add_argument('--queue', default='tornado_redis',
    help='Queue type')
argparser.add_argument('--redis-host', default='localhost',
    help='Redis hostname')
argparser.add_argument('--redis-port', default=6379, type=int,
    help='Redis port')
argparser.add_argument('--publish-method',
    help='If publishing messages to the queue is allowed from tornadopush')
argparser.add_argument('--webhook-url',
    help='A URL to POST data to when a message is received from the client')

def main():
    args = argparser.parse_args()
    settings = {
        "queue": args.queue,
        "token_secret": args.secret,
        "auth_enabled": args.auth,
        "token_max_age": args.tokenttl,
        "default_channel": args.default_channel,
        "cache_path": args.assets_cache_path,
        "redis_host": args.redis_host,
        "redis_port": args.redis_port,
        "publish_method": args.publish_method,
        "webhook_url": args.webhook_url,
        "debug": args.debug
    }
    if args.debug:
        logger.setLevel(logging.DEBUG)
    if os.path.exists(args.config):
        with open(args.config) as f:
            settings.update(yaml.load(f.read()))
    print "Starting tornadopush on port %s" % args.port
    print "Press Ctrl+C to exit"
    signal.signal(signal.SIGINT, signal_handler)
    tornado.ioloop.PeriodicCallback(try_exit, 100).start()
    start_server(args.port, **settings)


if __name__ == '__main__':
    main()